local floor = math.floor
local function compactify(n)
	local suffix = 1
	local new
	while n >= 1000 do
		new = floor(n / 100) / 10
		if n == new then
			return {'big-numbers.infinity'}
		else
			n = new
		end
		suffix = suffix + 1
	end
	return {'big-numbers.' .. suffix, n}
end

local power_usages = {
	['0W'] = 0,
	['60kW'] = 1000,
	['180kW'] = 3000,
	['300kW'] = 5000,
	['480kW'] = 8000,
	['600kW'] = 10000,
	['1.2MW'] = 20000,
	['2.4MW'] = 40000
}

local combinator_shift_x = 2.25
local combinator_shift_y = 1.75

local unit_types = {
	['memory-unit'] = true,
	['active-provider-memory-unit'] = true,
	['passive-provider-memory-unit'] = true,
	['storage-memory-unit'] = true,
	['requester-memory-unit'] = true,
	['buffer-memory-unit'] = true
}

local function setup()
	global.units = global.units or {}
end

script.on_init(setup)
script.on_configuration_changed(setup)

local function on_created(event)
	local entity = event.created_entity or event.entity
	local position = entity.position
	local surface = entity.surface
	
	if entity.type == 'entity-ghost' and entity.ghost_name == 'memory-unit-combinator' then
		if surface.count_entities_filtered{name = 'memory-unit-combinator', position = entity.position} ~= 0 then
			entity.destroy()
			return
		end
		local x = position.x - combinator_shift_x
		local y = position.y - combinator_shift_y
		for _, unit in pairs(surface.find_entities_filtered{type = 'entity-ghost', position = {x, y}}) do
			if unit_types[unit.ghost_name] and unit.position.x == x and unit.position.y == y then
				_, entity = entity.revive()
				entity.operable = false
				entity.destructible = false
				return
			end
		end
		entity.destroy()
	elseif unit_types[entity.name] then
		local force = entity.force
		
		local combinator_position = {position.x + combinator_shift_x, position.y + combinator_shift_y}
		local combinator = surface.find_entity('memory-unit-combinator', combinator_position)
		if not combinator then
			combinator = surface.create_entity{
				name = 'memory-unit-combinator',
				position = combinator_position,
				force = force
			}
			combinator.operable = false
			combinator.destructible = false
		end
		
		local powersource = surface.create_entity{
			name = 'memory-unit-powersource',
			position = position,
			force = force
		}
		powersource.destructible = false
		
		global.units[entity.unit_number] = {
			entity = entity,
			count = 0,
			powersource = powersource,
			combinator = combinator,
			inventory = entity.get_inventory(defines.inventory.chest)
		}
	end
end

script.on_event(defines.events.on_built_entity, on_created)
script.on_event(defines.events.on_robot_built_entity, on_created)
script.on_event(defines.events.script_raised_built, on_created)
script.on_event(defines.events.script_raised_revive, on_created)

script.on_event(defines.events.on_entity_cloned, function(event)
	local source = event.source	
	if not unit_types[source.name] then return end
	
	local destination = event.destination
	local unit_data = global.units[source.unit_number]
	local position = destination.position
	local surface = destination.surface
	
	local combinator, powersource = unit_data.combinator, unit_data.powersource
	combinator_position = {position.x + combinator_shift_x, position.y + combinator_shift_y}
	if combinator.valid then
		combinator = combinator.clone{position = combinator_position, surface = surface}
	else
		combinator = surface.create_entity{
			name = 'memory-unit-combinator',
			position = combinator_position,
			force = force
		}
		combinator.operable = false
		combinator.destructible = false
	end
	
	if powersource.valid then
		powersource = powersource.clone{position = position, surface = surface}
	else
		powersource = surface.create_entity{
			name = 'memory-unit-powersource',
			position = position,
			force = force
		}
		powersource.destructible = false
	end
	
	global.units[destination.unit_number] = {
		inventory = destination.get_inventory(defines.inventory.chest),
		powersource = powersource,
		combinator = combinator,
		item = unit_data.item,
		count = unit_data.count,
		entity = destination,
		comfortable = unit_data.comfortable,
		stack_size = unit_data.stack_size
	}
end)

local function on_destroyed(event)
	local entity = event.entity
	if entity.type == 'entity-ghost' and unit_types[entity.ghost_name] then
		local position = entity.position
		local combinator = entity.surface.find_entity('memory-unit-combinator', {position.x + combinator_shift_x, position.y + combinator_shift_y})
		if combinator then combinator.destroy() end
	elseif unit_types[entity.name] then
		local unit_data = global.units[entity.unit_number]
		global.units[entity.unit_number] = nil
		unit_data.powersource.destroy()
		unit_data.combinator.destroy()
		
		local buffer = event.buffer
		local item = unit_data.item
		local count = unit_data.count
		
		if buffer and item and count ~= 0 then
			buffer.insert{name = 'memory-element', count = 1}
			local stack = buffer.find_item_stack('memory-element')
			stack.tags = {name = item, count = count}
			stack.custom_description = {'item-description.memory-element-active', compactify(count), item}
		end
		
		local map_tag = unit_data.map_tag
		if map_tag and map_tag.valid then
			map_tag.destroy()
		end
	end
end

script.on_event(defines.events.on_player_mined_entity, on_destroyed)
script.on_event(defines.events.on_robot_mined_entity, on_destroyed)
script.on_event(defines.events.on_entity_died, on_destroyed)
script.on_event(defines.events.script_raised_destroy, on_destroyed)

local function pre_mined(event)
	local entity = event.entity
	if not unit_types[entity.name] then return end
	
	local unit_data = global.units[entity.unit_number]
	local item = unit_data.item
	
	if item then
		local inventory = unit_data.inventory
		local in_inventory = inventory.get_item_count(item)
		
		if in_inventory > 0 then
			unit_data.count = unit_data.count + inventory.remove{name = item, count = in_inventory}
		end
	end
end

script.on_event(defines.events.on_pre_player_mined_item, pre_mined)
script.on_event(defines.events.on_robot_pre_mined, pre_mined)
script.on_event(defines.events.on_marked_for_deconstruction, pre_mined)

local update_rate = 1 * 60
script.on_nth_tick(update_rate, function()
	for unit_number, unit_data in pairs(global.units) do
		local entity = unit_data.entity
		local powersource = unit_data.powersource
		local combinator = unit_data.combinator
		
		if entity.valid == false or powersource.valid == false or combinator.valid == false then
			if powersource.valid then
				powersource.destroy()
			end
			if combinator.valid then
				combinator.destroy()
			end
			game.print{'memory-unit-corruption', unit_data.count, unit_data.item or 'nothing'}
			global.units[unit_number] = nil
			goto next
		end
		
		local energy = powersource.energy
		if powersource.electric_buffer_size ~= energy then
			if energy ~= 0 then
				rendering.draw_sprite{
					sprite = 'utility.electricity_icon', 
					x_scale = 0.5,
					y_scale = 0.5,
					target = entity, 
					surface = entity.surface,
					time_to_live = update_rate / 2
				}
			end
			goto next
		end
		if entity.to_be_deconstructed() then goto next end
		
		local item = unit_data.item
		local inventory = unit_data.inventory
		local changed = false
		local sort = true
		
		if item == nil then
			for i = 1, #inventory do
				local stack = inventory[i]
				if stack.valid_for_read then
					local name = stack.name
					if name == 'memory-element' then
						local name = stack.tags.name
						if name and game.item_prototypes[name] then
							local tags = stack.tags
							item = tags.name
							unit_data.count = tags.count
							stack.clear()
						else
							goto bad
						end
					elseif name == 'empty-memory-element' then
						goto bad
					else
						item = name
						unit_data.count = 0
					end
					unit_data.item = item
					unit_data.stack_size = game.item_prototypes[item].stack_size
					unit_data.comfortable = unit_data.stack_size * entity.prototype.get_inventory_size(defines.inventory.chest) / 2
					changed = true
					
					unit_data.map_tag = entity.force.add_chart_tag(entity.surface, {
						position = entity.position,
						icon = {type = 'item', name = item}
					})
					
					break
					::bad::
				end
			end
		end
		
		if item == nil then goto next end
	
		local comfortable = unit_data.comfortable
		local inventory_count = inventory.get_item_count(item)
		
		for i = 1, #inventory do
			local stack = inventory[i]
			if stack.valid_for_read then
				local name = stack.name
				if name == 'memory-element' then
					local tags = stack.tags
					if tags.name == item then
						unit_data.count = unit_data.count + tags.count
						stack.clear()
						changed = true
					end
				else
					local total_count = inventory_count + unit_data.count
					if total_count ~= 0 and name == 'empty-memory-element' then
						stack.set_stack{name = 'memory-element', count = 1}
						stack.tags = {name = item, count = total_count}
						stack.custom_description = {'item-description.memory-element-active', compactify(total_count), item}
						
						inventory.remove{name = item, count = inventory_count}
						inventory_count = 0
						unit_data.count = 0
						
						sort = false
						changed = true
					end
				end
			end
		end
		
		if inventory_count > comfortable then
			local amount_removed = inventory.remove{name = item, count = inventory_count - comfortable}
			
			unit_data.count = unit_data.count + amount_removed
			inventory_count = inventory_count - amount_removed
			changed = true
		elseif inventory_count < comfortable then
			local to_add = comfortable - inventory_count
			if unit_data.count < to_add then
				to_add = unit_data.count
			end
			if to_add ~= 0 then
				local amount_added = inventory.insert{name = item, count = to_add}
				
				unit_data.count = unit_data.count - amount_added
				inventory_count = inventory_count + amount_added
			end
			changed = true
		end
		
		if changed == false then goto next end
		
		local total_count = unit_data.count + inventory_count
		
		combinator.get_or_create_control_behavior().set_signal(1, {
			signal = {type = 'item', name = item},
			count = math.min(2147483647, total_count)
		})
	
		if unit_data.text then
			rendering.set_text(unit_data.text, compactify(total_count))
		else
			unit_data.text = rendering.draw_text{
				surface = entity.surface,
				target = entity,
				text = compactify(total_count),
				alignment = 'center',
				scale = 1.5,
				only_in_alt_mode = true,
				color = {r = 1, g = 1, b = 1}
			}
		end
		
		if sort then
			inventory.sort_and_merge()
		end
		
		local power_usage = 1000 + (math.ceil(total_count / unit_data.stack_size) ^ 0.4) * power_usages[settings.global['memory-unit-power-usage'].value]
		powersource.power_usage = power_usage
		powersource.electric_buffer_size = power_usage
		
		if inventory_count == 0 then
			unit_data.alert = true
			for _, player in pairs(game.players) do
				if player.force == entity.force then
					player.add_custom_alert(entity, {type = 'item', name = 'memory-unit'}, {'memory-unit-empty', entity.localised_name, item}, true)
				end
			end
		elseif unit_data.alert then
			unit_data.alert = nil
			for _, player in pairs(game.players) do
				if player.force == entity.force then
					player.remove_alert{entity = entity}
				end
			end
		end
		
		::next::
	end
end)

remote.add_interface('memory-storage', {
	['give-memory-element'] = function(item, count, player)
		player = player or game.player
		if player == nil then error('You need to pass a player since you are not using this from the console') end
		
		if item == nil then player.print{'command-output.bad-args'} return end
		local type
		if game.item_prototypes[item] then
			type = 'item'
		elseif game.fluid_prototypes[item] and script.active_mods['fluid-memory-storage'] then
			type = 'fluid'
		else
			player.print{'command-output.bad-item', item}
			return
		end
		
		if count == nil then player.print{'command-output.bad-args'} return end
		if count <= 0 then player.print{'command-output.bad-count'} return end
		
		local inventory = player.get_main_inventory()
		inventory.insert{name = 'memory-element', count = 1}
		for i = 1, #inventory do
			local stack = inventory[i]
			if stack.valid_for_read and stack.name == 'memory-element' then
				local tags = stack.tags
				if next(tags) == nil then
					if type == 'item' then
						stack.tags = {name = item, count = count}
						stack.custom_description = {
							'item-description.memory-element-active',
							compactify(count),
							item
						}
					else
						local temperature = game.fluid_prototypes[item].default_temperature
						stack.tags = {name = item, count = count, temperature = temperature}
						stack.custom_description = {
							'item-description.fluid-memory-element-active',
							compactify(count),
							item,
							temperature
						}
					end
					break
				end
			end
		end
	end
})